#include "stationplanmodel.h"

#include "utils/platform_utils.h"
#include "utils/jobcategorystrings.h"

#include <QColor>

StationPlanModel::StationPlanModel(sqlite3pp::database &db, QObject *parent) :
    QAbstractTableModel(parent),
    mDb(db),
    q_countPlanItems(mDb, "SELECT COUNT(id) FROM stops WHERE stationId=?"),
    q_selectPlan(mDb, "SELECT stops.id,"
                      "jobs.category,"
                      "jobs.id,"
                      "stops.arrival,"
                      "stops.departure,"
                      "stops.platform,"
                      "stops.transit"
                      " FROM stops"
                      " JOIN jobs ON jobs.id=stops.jobId"
                      " WHERE stops.stationId=?"
                      " ORDER BY stops.arrival,stops.platform")
{
}

QVariant StationPlanModel::headerData(int section, Qt::Orientation orientation, int role) const
{
    if(orientation == Qt::Horizontal && role == Qt::DisplayRole)
    {
        switch (section)
        {
        case Arrival:
            return tr("Arrival");
        case Departure:
            return tr("Departure");
        case Platform:
            return tr("Platform");
        case Job:
            return tr("Job");
        case Notes:
            return tr("Notes");
        default:
            break;
        }
    }
    return QAbstractTableModel::headerData(section, orientation, role);
}

int StationPlanModel::rowCount(const QModelIndex &parent) const
{
    return parent.isValid() ? 0 : m_data.size();
}

int StationPlanModel::columnCount(const QModelIndex &parent) const
{
    return parent.isValid() ? 0 : NCols;
}

QVariant StationPlanModel::data(const QModelIndex &idx, int role) const
{
    if (!idx.isValid() || idx.row() >= m_data.size() || idx.column() >= NCols)
        return QVariant();

    const StPlanItem& item = m_data.at(idx.row());

    switch (role)
    {
    case Qt::DisplayRole:
    {

        switch (idx.column())
        {
        case Arrival:
        {
            if(item.type == StPlanItem::ItemType::Departure)
                break; //Don't repeat arrival also in the second row.
            return item.arrival;
        }
        case Departure:
            return item.departure;
        case Platform:
            return utils::platformName(item.platform);
        case Job:
            return JobCategoryName::jobName(item.jobId, item.cat);
        case Notes:
        {
            if(item.type == StPlanItem::ItemType::Departure)
                return StationPlanModel::tr("Departure"); //Don't repeat description also in the second row.

            return item.description;
        }
        }
        break;
    }
    case Qt::ForegroundRole:
    {
        if(item.type == StPlanItem::ItemType::Transit)
            return QColor(0, 0, 255); //Transit in blue
        break;
    }
    case Qt::ToolTipRole:
    {
        if(item.type == StPlanItem::ItemType::Transit)
            return StationPlanModel::tr("Transit");
        break;
    }
    }

    return QVariant();
}

void StationPlanModel::clear()
{
    beginResetModel();
    m_data.clear();
    m_data.squeeze();
    endResetModel();
}

void StationPlanModel::loadPlan(db_id stId)
{
    beginResetModel();

    m_data.clear();

    q_countPlanItems.bind(1, stId);
    q_countPlanItems.step();
    int count = q_countPlanItems.getRows().get<int>(0);
    q_countPlanItems.reset();

    if(count > 0)
    {
        QMap<QTime, StPlanItem> stopMap; //Order by Departure ASC

        q_selectPlan.bind(1, stId);
        for(auto r : q_selectPlan)
        {
            StPlanItem curStop;
            curStop.stopId = r.get<db_id>(0);
            curStop.cat = JobCategory(r.get<int>(1));
            curStop.jobId = r.get<db_id>(2);
            curStop.arrival = r.get<QTime>(3);
            curStop.departure = r.get<QTime>(4);
            curStop.platform = r.get<int>(5);
            curStop.type = StPlanItem::ItemType::Normal;
            int stopType = r.get<int>(6);

            for(auto stop = stopMap.begin(); stop != stopMap.end(); /*nothing because of erase */)
            {
                if(stop->departure <= curStop.arrival)
                {
                    m_data.append(stop.value());
                    m_data.last().type = StPlanItem::ItemType::Departure;
                    stop = stopMap.erase(stop);
                }
                else
                {
                    ++stop;
                }
            }

            m_data.append(curStop);
            if(stopType == Transit)
            {
                m_data.last().type = StPlanItem::ItemType::Transit;
            }
            else
            {
                //Transit should not be repeated
                curStop.type = StPlanItem::ItemType::Departure;
                stopMap.insert(curStop.departure, std::move(curStop));
            }
        }
        q_selectPlan.reset();

        for(const StPlanItem& stop : stopMap)
        {
            m_data.append(stop);
        }
    }

    m_data.squeeze();

    endResetModel();
}

std::pair<db_id, db_id> StationPlanModel::getJobAndStopId(int row) const
{
    if(row < m_data.size())
    {
        const StPlanItem& item = m_data.at(row);
        return {item.jobId, item.stopId};
    }
    return {0, 0};
}
